﻿/*
Copyright (c) 2013, Douglas Stebila, Brendon Park, David Sweeney, Nathan Wills, Simon Lockhart, Thomas Phan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice, this
  list of conditions and the following disclaimer in the documentation and/or
  other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography;
using System.Security.Policy;
using System.Text;
using System.Threading.Tasks;

namespace NMLE
{
    /// <summary>
    ///     Exception thrown when the output hash size of the supplied 
    ///     hash function is incompatable with the legal key sizes of the 
    ///     supplied symetric algorythm.
    /// </summary>
    public class IncompatableAlgorithmException : Exception
    {
        /// <summary>
        ///     Gets the size of the hash the supplied hash function ouputs
        /// </summary>
        public int HashKeySize { get; private set; }

        /// <summary>
        ///     Gets the KeySizes object of the supplied symetric algorythm
        /// </summary>
        public KeySizes[] EncryptionKeySizes { get; private set; }

        /// <summary>
        ///     Initaliases a new instance of the IncompatableAlgorithmException 
        ///     class.
        /// </summary>
        /// <param name="hashAlgorithm">
        ///     The attempted HashAlgorythm
        /// </param>
        /// <param name="encryptionAlgorithm">
        ///     The attempted SymetricAlgorytm
        /// </param>
        public IncompatableAlgorithmException(
            HashAlgorithm hashAlgorithm,
            SymmetricAlgorithm encryptionAlgorithm
            ): base(ExceptionMessage(hashAlgorithm, encryptionAlgorithm))
        {
            EncryptionKeySizes = encryptionAlgorithm.LegalKeySizes;
            HashKeySize = hashAlgorithm.HashSize;
        }

        private static string ExceptionMessage(
            HashAlgorithm hashAlgorithm,
            SymmetricAlgorithm encryptionAlgorithm
            )
        {
            return String.Format(
                "The supplied hash algorythm does not produce a hash of " +
                "adequate size ({0}) to suit the supplied encryption " +
                "algorytm ({1})",
                hashAlgorithm.HashSize,
                PrintValidKeySizes(encryptionAlgorithm.LegalKeySizes)
                );
        }

        private static string PrintValidKeySizes(IEnumerable<KeySizes> keySizes)
        {
            var output = new StringBuilder();
            foreach (var keySize in keySizes)
                for (int i = keySize.MinSize;
                    i <= keySize.MaxSize;
                    i = i + keySize.SkipSize)
                    output.Append(i + ", ");
            return output.ToString();
        }
    }
}
