﻿/*
Copyright (c) 2013, Douglas Stebila, Brendon Park, David Sweeney, Nathan Wills, Simon Lockhart, Thomas Phan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice, this
  list of conditions and the following disclaimer in the documentation and/or
  other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;
using System.Security.Cryptography;

namespace NMLE.RCE
{
    /// <summary>
    /// 
    /// </summary>
    /// <typeparam name="THashAlgorithm"></typeparam>
    /// <typeparam name="TEncryptionAlgorithm"></typeparam>
    public class RCETransform<THashAlgorithm, TEncryptionAlgorithm>
        : MLETransformBase<THashAlgorithm, TEncryptionAlgorithm>
        where THashAlgorithm : HashAlgorithm, new()
        where TEncryptionAlgorithm : SymmetricAlgorithm, new()
    {

        private Byte[] _key;

        /// <summary>
        ///     Gets the Key of the algorytm
        /// </summary>
        public override byte[] Key
        {
            get { return _key; }
        }

        /// <summary>
        ///     The lockValue used to perform the cryptographic transform
        /// </summary>
        public byte[] LockValue
        {
            get { return TransformKey; }
        }

        /// <summary>
        ///     The crypto value that stores the key and lockValue
        /// </summary>
        public byte[] CryptoValue
        {
            get { return XORArraBytes(LockValue, Key); }
        }

        /// <summary>
        ///     Initalises a new RCE transform
        /// </summary>
        /// <param name="lockValue">
        ///     The lockValue used to transform the message
        /// </param>
        /// <param name="direction">
        ///     The direction of encyption
        /// </param>
        public RCETransform(
            Byte[] lockValue,
            EncryptionDirection direction
            )
            : base(lockValue, direction)
        {
        }

        /// <summary>
        ///     Initalises a new RCE transform
        /// </summary>
        /// <param name="key">
        ///     The key used to encrypt the transform
        /// </param>
        /// <param name="c2Value">
        ///     The crypto value holding the key and lock value
        /// </param>
        /// <param name="direction">
        ///     The direction of encyption
        /// </param>
        public RCETransform(
            Byte[] key,
            Byte[] c2Value,
            EncryptionDirection direction
            )
            : base(XORArraBytes(c2Value,key), direction)
        {
            _key = key;
        }

        /// <summary>
        ///     Transforms the specified region of the input byte array and
        ///     copies the resulting transform to the specified region of the
        ///     output byte array.
        /// </summary>
        /// <returns>
        ///     The number of bytes written.
        /// </returns>
        /// <param name="inputBuffer">
        ///     The input for which to compute the transform.
        /// </param>
        /// <param name="inputOffset">
        ///     The offset into the input byte array from which to begin using
        ///     data.
        /// </param>
        /// <param name="inputCount">
        ///     The number of bytes in the input byte array to use as data.
        /// </param>
        /// <param name="outputBuffer">
        ///     The output to which to write the transform.
        /// </param>
        /// <param name="outputOffset">
        ///     The offset into the output byte array from which to begin
        ///     writing data.
        /// </param>
        public override int TransformBlock(
            byte[] inputBuffer,
            int inputOffset,
            int inputCount,
            byte[] outputBuffer,
            int outputOffset
            )
        {
            var result = UnderlyingSymetricTransform.TransformBlock(
                inputBuffer,
                inputOffset,
                inputCount,
                outputBuffer,
                outputOffset
                );

            var hashBuffer = Direction == EncryptionDirection.Decrypt
                ? outputBuffer
                : inputBuffer;

            UnderlyingHashTransform.TransformBlock(
                hashBuffer,
                inputOffset,
                inputCount,
                hashBuffer,
                0
                );

            return result;
        }

        /// <summary>
        ///     Transforms the specified region of the specified byte array.
        /// </summary>
        /// <returns>
        ///     The computed transform.
        /// </returns>
        /// <param name="inputBuffer">
        ///     The input for which to compute the transform. 
        /// </param>
        /// <param name="inputOffset">
        ///     The offset into the byte array from which to begin using data. 
        /// </param>
        /// <param name="inputCount">
        ///     The number of bytes in the byte array to use as data. 
        /// </param>
        public override byte[] TransformFinalBlock(
            byte[] inputBuffer,
            int inputOffset,
            int inputCount
            )
        {
            var lastBytes = UnderlyingSymetricTransform.TransformFinalBlock(
                inputBuffer,
                inputOffset,
                inputCount
                );

            var hashBuffer = Direction == EncryptionDirection.Decrypt
                ? lastBytes
                : inputBuffer;

            UnderlyingHashTransform.TransformFinalBlock(
                hashBuffer,
                inputOffset,
                inputCount
                );
            
            if (Direction == EncryptionDirection.Decrypt)
                ReportHashCheck(UnderlyingHashTransform.Hash);
            else
                _key = UnderlyingHashTransform.Hash;

            return lastBytes;
        }



        private static byte[] XORArraBytes(byte[] buffer1, byte[] buffer2)
        {
            for (int i = 0; i < buffer1.Length; i++)
                buffer1[i] ^= buffer2[i];
            return buffer1;
        }

    }
}